--------------------------------------------------------
-- Minetest :: ActiveFormspecs Mod (formspecs)
--
-- See README.txt for licensing and release notes.
-- Copyright (c) 2016-2020, Leslie E. Krause
--
-- ./games/minetest_game/mods/formspecs/init.lua
--------------------------------------------------------

core.FORMSPEC_SIGEXIT = "true"		-- player pressed esc key
core.FORMSPEC_SIGQUIT = 1		-- player logged off
core.FORMSPEC_SIGTERM = 3		-- client is shutting down
core.FORMSPEC_SIGPROC = 4		-- procedural closure
core.FORMSPEC_SIGTIME = 5		-- timeout reached
core.FORMSPEC_SIGSTOP = 6		-- procedural closure (cannot be trapped)
core.FORMSPEC_SIGHOLD = 7		-- child form opened, parent is suspended
core.FORMSPEC_SIGCONT = 8		-- child form closed, parent can continue

-----------------------------------------------------------------

local afs = { }		-- obtain localized, protected namespace

afs.menu = nil
afs.timers = { }


local _ = { }

local function is_match( str, pat )
	-- use array for captures
	_ = { string.match( str, pat ) }
	return #_ > 0 and _ or nil
end

local function recolor( text )
	return string.gsub( text, "@(#.-){(.-)}", function( color, input )
		return core.colorize( color, input )
	end )
end

local function escape( str, pat )
	return string.gsub( str, pat, function ( c )
		return string.format( "\\x%00X", string.byte( c ) )
	end )
end

local function unescape( str )
	return string.gsub( str, "\\x(..)", function ( b )
		return "\\" .. string.char( tonumber( b, 16 ) )
	end )
end

local function unescape_raw( str )
	return string.gsub( str, "\\x(..)", function ( b )
		return string.char( tonumber( b, 16 ) )
	end )
end

-----------------------------------------------------------------
-- trigger callbacks at set intervals within timer queue
-----------------------------------------------------------------

do
	-- localize needed object references for efficiency
	local clock = core.get_us_time
	local rtime = clock( ) / 1000000

	afs.get_uptime = function( )
		return clock( ) / 1000000 - rtime
	end

	afs.clear_timers = function( )
		timers = { }
		afs.timers = timers
	end

--[[	core.register_globalstep( function ( )
		local uptime = clock( ) / 1000000 - rtime

		if afs.timer and uptime >= afs.timer.exptime then
			local timer = afs.timer

			timer.counter = self.counter + 1
			timer.overrun = uptime - timer.exptime
			timer.exptime = uptime + timer.menu.timeout

			timer.menu.newtime = uptime
			timer.menu.on_close( timer.menu.meta, { quit = core.FORMSPEC_SIGTIME } )

			timer.overrun = 0.0
		end
	end ) ]]
end

-----------------------------------------------------------------
-- define standard formspec variables and functions
-----------------------------------------------------------------

core.gui_bg = "bgcolor[#080808BB;true]"
core.gui_bg_img = "background[5.0,5.0;1.0,1.0;gui_formbg.png;true]"

-----------------------------------------------------------------
-- trigger callbacks during form-related events
-----------------------------------------------------------------

core.button_handler = function ( fields )
	local menu = afs.menu

	-- handle reverse-lookups of dropdown fields
	for name, keys in pairs( menu.dropdowns ) do
		if fields[ name ] then
			fields[ name ] = keys[ fields[ name ] ]
		end
	end

	menu.newtime = afs.get_uptime( )
	menu.on_close( menu.meta, fields )
end

-----------------------------------------------------------------
-- convert contextual fields to key-value pairs
-----------------------------------------------------------------

core.get_context_field = function ( fields, prefix )
	for k, v in pairs( fields ) do
		if is_match( k, "^(" .. prefix .. "):(.+)$" ) then
			return _[ 2 ], _[ 1 ]
		end
	end
end

-----------------------------------------------------------------
-- expose timer functionality within a helper object
-----------------------------------------------------------------

core.get_menu_timer = function ( )
	local self = { }
	local menu = afs.menu

	self.start = function ( timeout )
		if not menu.timeout and timeout >= 0.1 then
			local uptime = afs.get_uptime( )

			menu.timeout = timeout
			afs.timer = { menu = menu, counter = 0, oldtime = uptime, exptime = uptime + timeout, overrun = 0.0 }
		end
	end

	self.stop = function ( )
		if not menu.timeout then return end

		menu.timeout = nil
		afs.timer = nil
	end

	self.get_state = function ( )
		local uptime = afs.get_uptime( )

		if afs.timer then
			return { elapsed = uptime - afs.timer.oldtime, remain = afs.timer.exptime - uptime, overrun = afs.timer.overrun, counter = afs.timer.counter }
		end

	end

	return self
end

-----------------------------------------------------------------
-- construct object to manage state of formspec groups
-----------------------------------------------------------------

local function create_groups( formspec )
	local self = { }
	local groups = { }

	local function tokenize( str )
		str = escape( str, "[{}]" )
		str = string.gsub( str, "group%[(.-)%]", "%1:{" )
		str = string.gsub( str, "group_end%[%]", "}" )
		return str
	end

	local function insert_raw( name, text )
		text = string.gsub( text, "([a-zA-Z_]+):(%b{})", function ( subname, subtext )
			insert_raw( subname, string.sub( subtext, 2, -2 ) )
			return "group[" .. subname .. "]"
		end )
		groups[ name ] = text
	end

	self.insert = function ( name, formspec )
		insert_raw( name, tokenize( formspec ) )
	end

	self.update = function ( name, formspec )
		for subname in string.gmatch( groups[ name ], "group%[(.-)%]" ) do
			self.update( subname )
		end
		if formspec then
			self.insert( name, formspec )
		else
			groups[ name ] = nil
		end
	end

	self.compile = function ( name )
		local text = string.gsub( groups[ name or "root" ], "group%[(.-)%]", function ( subname )
			return self.compile( subname )
		end )
		return unescape_raw( text )
	end

	self.insert( "root", formspec )

	return self
end

-----------------------------------------------------------------
-- parse specialized formspec elements and escapes codes
-----------------------------------------------------------------

local function parse_elements( menu, formspec )
	formspec = escape( formspec, "\\([,;%[%]])" )
	menu.dropdowns = { }	-- reset the dropdown lookup

	-- dropdown elements can optionally return the selected
	-- index rather than the value of the option itself
	formspec = string.gsub( formspec, "dropdown%[(.-)%]", function( params )
		if is_match( params, "^([^;]*;[^;]*;([^;]*);([^;]*);[^;]*);([^;]*)$" ) then
			local prefix = _[ 1 ]
			local name = _[ 2 ]
			local options = _[ 3 ]
			local use_index = _[ 4 ]

			if use_index == "true" then
				menu.dropdowns[ name ] = { }
				for idx, val in ipairs( string.split( options, ",", true ) ) do
					menu.dropdowns[ name ][ unescape_raw( val ) ] = idx     -- add to reverse lookup table
				end
				return string.format( "dropdown[%s]", prefix )
			elseif use_index == "false" or use_index == "" then
				return string.format( "dropdown[%s]", prefix )
			else
				return ""	-- strip invalid dropdown elements
			end
		end
		return string.format( "dropdown[%s]", params )
	end )

	-- hidden elements only provide default, initial values 
	-- for state table and are always stripped afterward
	formspec = string.gsub( formspec, "hidden%[(.-)%]", function( params )
		if is_match( params, "^([^;]*);([^;]*)$" ) or is_match( params, "^([^;]*);([^;]*);([^;]*)$" ) then
			local key = _[ 1 ]
			local value = _[ 2 ]
			local type = _[ 3 ]

			if key ~= "" and menu.meta[ key ] == nil then
				-- parse according to specified data type
				if type == "string" or type == "" or type == nil then
					menu.meta[ key ] = unescape_raw( value )
				elseif type == "number" then
					menu.meta[ key ] = tonumber( value )
				elseif type == "boolean" then
					menu.meta[ key ] = ( { ["1"] = true, ["0"] = false, ["true"] = true, ["false"] = false } )[ value ]
				end
			end
		end
		return ""       -- strip hidden elements prior to showing formspec
	end )

	return unescape( formspec )
end

-----------------------------------------------------------------
-- open detached formspec with session-based state table
-----------------------------------------------------------------

core.create_menu = function ( meta, formspec, on_close, signal )
	-- short circuit whenever required params are missing
	if not formspec then return end

	local menu = afs.menu

	-- trigger previous callback before formspec closure
	if menu then
		core.get_menu_timer( ).stop( )
		if signal ~= core.FORMSPEC_SIGSTOP then
			menu.on_close( menu.meta, { quit = signal or core.FORMSPEC_SIGPROC } )
		end
		if signal ~= core.FORMSPEC_SIGHOLD then
			menu = menu.parent_menu
		end
	end

	menu = { parent_menu = menu }
	menu.on_close = on_close or function ( ) end
	menu.meta = meta or { }
	menu.oldtime = afs.get_uptime( )
	menu.newtime = menu.oldtime
	menu.count = 1
	menu.groups = create_groups( formspec )

	afs.menu = menu
	core.update_formspec( parse_elements( menu, menu.groups.compile( ) ) )
end

core.update_menu = function ( formspec )
	local menu = afs.menu

	if menu then
		menu.oldtime = afs.get_uptime( )
		menu.count = menu.count + 1

		if type( formspec ) == "table" then
			for subname, subtext in pairs( formspec ) do
				menu.groups.update( subname, subtext )
			end
			core.update_formspec( parse_elements( menu, menu.groups.compile( ) ) )
		else
			menu.groups.update( "root", formspec )
			core.update_formspec( parse_elements( menu, formspec ) )
		end
	end
end

core.destroy_menu = function ( signal )
	local menu = afs.menu

	if menu then
		if signal ~= core.FORMSPEC_SIGSTOP then
			menu.newtime = afs.get_uptime( )
			menu.on_close( menu.meta, { quit = signal or core.FORMSPEC_SIGPROC } )
		end

		core.get_menu_timer( ).stop( )

		if menu.parent_menu then
			-- restore previous session
			menu = menu.parent_menu
			menu.newtime = afs.get_uptime( )
			afs.menu = menu

			menu.count = 1
			menu.on_close( menu.meta, { quit = core.FORMSPEC_SIGCONT } )

			if menu.count == 1 then
				-- purge stale formspec and forcibly end session
				core.close( )
			end
		else
			-- end current session
			core.close( )
		end
	end
end